from transifex import native
from transifex.native import tx as _tx
from transifex.native.rendering import PseudoTranslationPolicy


class TestModuleInit(object):
    """Test __init__.py of root native module."""

    def test_init_uses_given_params(self, reset_tx):
        native.init(
            'mytoken', ['lang1', 'lang2'], cds_host='myhost',
            missing_policy=PseudoTranslationPolicy(),
        )
        assert _tx.initialized is True
        assert _tx._cds_handler.token == 'mytoken'
        assert _tx._cds_handler.host == 'myhost'
        assert _tx._languages == ['lang1', 'lang2']
        assert isinstance(_tx._missing_policy, PseudoTranslationPolicy)

    def test_initialized_only_once(self, reset_tx):
        # Even if native.init() is called multiple times, only the first one matters
        native.init(
            'mytoken', ['lang1', 'lang2'], cds_host='myhost',
            missing_policy=PseudoTranslationPolicy(),
        )
        native.init('another_token', ['lang3', 'lang4'], 'another_host')
        assert _tx._cds_handler.token == 'mytoken'
        assert _tx._cds_handler.host == 'myhost'
        assert _tx._languages == ['lang1', 'lang2']
        assert isinstance(_tx._missing_policy, PseudoTranslationPolicy)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import mock
from django.core.management import call_command
from tests.native.django.test_tools.test_migrations.test_templatetags import (
    DJANGO_TEMPLATE, TRANSIFEX_TEMPLATE)
from transifex.native.django.management.commands.migratetransifex import \
    Command
from transifex.native.django.management.common import TranslatableFile
from transifex.native.parsing import SourceString
from transifex.native.tools.migrations.review import (FileReviewPolicy,
                                                      NoopReviewPolicy,
                                                      StringReviewPolicy)
from transifex.native.tools.migrations.save import (BackupSavePolicy,
                                                    InPlaceSavePolicy,
                                                    NewFileSavePolicy,
                                                    NoopSavePolicy)

PYTHON_TEMPLATE = u"""
# -*- coding: utf-8 -*-

{_import}

{call1}(u'{string1}', u'désign1,désign2', param1='1', param2=2, param3=True)
{call2}(
    u'{string2}', u'opération', _comment='comment', _tags='t1,t2', _charlimit=33,
)
"""

HTML_SAMPLE_1 = DJANGO_TEMPLATE
HTML_SAMPLE_2 = """
{% load i18n %}
{% load t from transifex %}

{% t "Hello!" %}

{% t "May" _context="month name" %}

{% t counter='something'|length name=name %}
{counter, plural, one {
There is only one {name} object.
} other {
There are {counter} {name} objects.
}}
{% endt %}

<a href="{{ url }}">Text</a>
"""
HTML_COMPILED_1 = TRANSIFEX_TEMPLATE

PATH_FIND_FILES = 'transifex.native.django.management.commands' \
                  '.migratetransifex.Command._find_files'
PATH_READ_FILE = 'transifex.native.django.management.commands' \
                 '.migratetransifex.Command._read_file'
PATH_PROMPT_FILE = 'transifex.native.tools.migrations.review.ReviewPolicy' \
                   '.prompt_for_file'
PATH_PROMPT_STRING = 'transifex.native.tools.migrations.review' \
                     '.ReviewPolicy.prompt_for_string'
PATH_SAVE_FILE = 'transifex.native.tools.migrations.save.SavePolicy' \
                 '._safe_save'
PATH_PROMPT_START1 = 'transifex.native.tools.migrations.review' \
                     '.prompt_to_start'
PATH_PROMPT_START2 = 'transifex.native.django.management.commands' \
                     '.migratetransifex.prompt_to_start'
PATH_ECHO = 'transifex.native.django.management.commands' \
    '.migratetransifex.Color.echo'


@mock.patch(PATH_ECHO)
@mock.patch(PATH_PROMPT_START2)
@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_dry_run_save_none_review(mock_find_files, mock_read,
                                  mock_prompt_to_start1,
                                  mock_prompt_to_start2, mock_echo):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
    ]
    command = Command()
    call_command(command, save_policy='dry-run', review_policy='none')
    assert isinstance(command.save_policy, NoopSavePolicy)
    assert isinstance(command.review_policy, NoopReviewPolicy)


@mock.patch(PATH_ECHO)
@mock.patch(PATH_PROMPT_START2)
@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_SAVE_FILE, return_value=(True, None))
@mock.patch(PATH_PROMPT_STRING)
@mock.patch(PATH_PROMPT_FILE)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_new_file_save_file_review(mock_find_files, mock_read,
                                   mock_prompt_file, mock_prompt_string,
                                   mock_save_file,
                                   mock_prompt_to_start1,
                                   mock_prompt_to_start2, mock_echo):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
    ]
    command = Command()
    call_command(command, save_policy='new', review_policy='file')
    assert isinstance(command.save_policy, NewFileSavePolicy)
    assert isinstance(command.review_policy, FileReviewPolicy)

    # The FileMigration instance that reached the review object
    # should compile to the proper Native syntax
    file_migration = mock_prompt_file.call_args[0][0]
    assert file_migration.compile() == HTML_COMPILED_1

    # No string review should have taken place
    assert mock_prompt_string.call_count == 0

    # The path and content that reached the save object
    # should have the correct values
    assert mock_save_file.call_args[0][0] == 'dir1/dir2/1__native.html'
    migration_compile = mock_save_file.call_args[0][1]
    assert migration_compile() == HTML_COMPILED_1


@mock.patch(PATH_ECHO)
@mock.patch(PATH_PROMPT_START2)
@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_SAVE_FILE, return_value=(True, None))
@mock.patch(PATH_PROMPT_STRING)
@mock.patch(PATH_PROMPT_FILE)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_backup_save_string_review(mock_find_files, mock_read,
                                   mock_prompt_file, mock_prompt_string,
                                   mock_save_file,
                                   mock_prompt_to_start1,
                                   mock_prompt_to_start2, mock_echo):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
    ]
    command = Command()
    call_command(command, save_policy='backup', review_policy='string')
    assert isinstance(command.save_policy, BackupSavePolicy)
    assert isinstance(command.review_policy, StringReviewPolicy)

    assert mock_prompt_file.call_count == 0
    assert mock_prompt_string.call_count == 7  # 7 migrated strings

    # The path and content that reached the save object
    # should have the correct values, one for the backup and one for
    # the new content
    assert mock_save_file.call_args_list[0][0][0] == 'dir1/dir2/1.html.bak'
    original_content_getter = mock_save_file.call_args_list[0][0][1]
    assert original_content_getter() == HTML_SAMPLE_1

    assert mock_save_file.call_args_list[1][0][0] == 'dir1/dir2/1.html'
    migration_compile = mock_save_file.call_args_list[1][0][1]
    assert migration_compile() == HTML_COMPILED_1


@mock.patch(PATH_ECHO)
@mock.patch(PATH_PROMPT_START2)
@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_SAVE_FILE, return_value=(True, None))
@mock.patch(PATH_PROMPT_STRING)
@mock.patch(PATH_PROMPT_FILE)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_replace_save_string_review(mock_find_files, mock_read,
                                    mock_prompt_file, mock_prompt_string,
                                    mock_save_file,
                                    mock_prompt_to_start1,
                                    mock_prompt_to_start2, mock_echo):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
    ]
    command = Command()
    call_command(command, save_policy='replace', review_policy='string')
    assert isinstance(command.save_policy, InPlaceSavePolicy)
    assert isinstance(command.review_policy, StringReviewPolicy)

    assert mock_prompt_file.call_count == 0
    assert mock_prompt_string.call_count == 7

    # The path and content that reached the save object
    # should have the correct values
    assert mock_save_file.call_args[0][0] == 'dir1/dir2/1.html'
    migration_compile = mock_save_file.call_args[0][1]
    assert migration_compile() == HTML_COMPILED_1



import sys

from mock import mock_open, patch
from transifex.native.tools.migrations.models import (FileMigration,
                                                      StringMigration)
from transifex.native.tools.migrations.save import (BackupSavePolicy,
                                                    InPlaceSavePolicy,
                                                    NewFileSavePolicy,
                                                    NoopSavePolicy, SavePolicy)

BUILTINS_MODULE = 'builtins' if sys.version_info >= (3, 0) else '__builtin__'


def _file_migration():
    migration = FileMigration('path/filename.html', 'the content')
    migration.add_string(StringMigration(
        'the content', 'the migrated content'))
    return migration


def test_noop_policy_does_not_open_file():
    policy = NoopSavePolicy()
    m = mock_open()
    with patch(BUILTINS_MODULE + ".open", m, create=True):
        saved, error_type = policy.save_file(_file_migration())
        assert saved is False
        assert error_type is None
    assert m.call_count == 0


def test_new_file_policy_writes_to_new_file():
    policy = NewFileSavePolicy()
    m = mock_open()
    with patch(BUILTINS_MODULE + ".open", m, create=True):
        saved, error_type = policy.save_file(_file_migration())
        assert saved is True
        assert error_type is None

    m.assert_called_once_with('path/filename__native.html', 'w')
    m().write.assert_called_once_with('the migrated content')


def test_backup_policy_writes_to_original_file_and_takes_backup():
    policy = BackupSavePolicy()
    m = mock_open()
    with patch(BUILTINS_MODULE + ".open", m, create=True):
        saved, error_type = policy.save_file(_file_migration())
        assert saved is True
        assert error_type is None

    assert m.call_args_list[0][0] == ('path/filename.html.bak', 'w')
    assert m.call_args_list[1][0] == ('path/filename.html', 'w')
    handler = m()
    assert handler.write.call_args_list[0][0] == (u'the content',)
    assert handler.write.call_args_list[1][0] == (u'the migrated content',)


def test_in_place_policy_writes_to_original_file():
    policy = InPlaceSavePolicy()
    m = mock_open()
    with patch(BUILTINS_MODULE + ".open", m, create=True):
        saved, error_type = policy.save_file(_file_migration())
        assert saved is True
        assert error_type is None

    m.assert_called_once_with('path/filename.html', 'w')
    m().write.assert_called_once_with('the migrated content')


@patch('transifex.native.tools.migrations.save.Color.echo')
def test_safe_save_handles_io_error(mock_echo):
    def raise_error():
        raise IOError()

    policy = SavePolicy()
    m = mock_open()
    with patch(BUILTINS_MODULE + ".open", m, create=True):
        saved, error_type = policy._safe_save(
            'doesnt-matter', raise_error, 'Dummy')
        assert saved is False
        assert error_type is IOError

    assert 'IOError while saving to dummy' in mock_echo.call_args[0][0]


@patch('transifex.native.tools.migrations.save.Color.echo')
def test_safe_save_handles_any_error(mock_echo):
    def raise_error():
        raise ValueError()

    policy = SavePolicy()
    m = mock_open()
    with patch(BUILTINS_MODULE + ".open", m, create=True):
        saved, error_type = policy._safe_save(
            'doesnt-matter', raise_error, 'Dummy')
        assert saved is False
        assert error_type is ValueError

    assert 'Error while saving to dummy' in mock_echo.call_args[0][0]



from django.conf import settings

TRANSIFEX_TOKEN = getattr(settings, 'TRANSIFEX_TOKEN', '')
TRANSIFEX_SECRET = getattr(settings, 'TRANSIFEX_SECRET', '')
TRANSIFEX_CDS_HOST = getattr(settings, 'TRANSIFEX_CDS_HOST', None)
TRANSIFEX_MISSING_POLICY = getattr(settings, 'TRANSIFEX_MISSING_POLICY', None)
TRANSIFEX_ERROR_POLICY = getattr(settings, 'TRANSIFEX_ERROR_POLICY', None)
LANGUAGES = getattr(settings, 'LANGUAGES', [])



import logging
import os
import sys

from django.apps import AppConfig
from django.core.signals import request_finished
from django.utils.translation import to_locale
from transifex.native import init, tx
from transifex.native.daemon import daemon
from transifex.native.django import settings as native_settings
from transifex.native.rendering import (parse_error_policy,
                                        parse_rendering_policy)

logger = logging.getLogger('transifex.native.django')
logger.addHandler(logging.StreamHandler(sys.stdout))


def _segments_match(segments_to_match, arguments):
    """
    Tries to find matching segments in the arguments of the command
    that started the app. Will be used to determine whether we want
    to fetch translations or not.

    :param list(str) segments_to_match: The segments that should be
    matched. Expects to match all segments and will match subsequences
    as well (e.g. `./manage.py` will match `manage.py`)
    :param list(str) arguments: The arguments of the command that started
    the app (probably `sys.argv`)
    :rtype bool: Whether segments match or not
    """

    segments_to_match = set(segments_to_match)
    for arg in arguments:
        for segment in list(segments_to_match):
            if segment in arg:
                segments_to_match.remove(segment)
                if not segments_to_match:
                    return True
    return not segments_to_match


class NativeConfig(AppConfig):

    name = "transifex.native.django"

    def ready(self):
        if not native_settings.TRANSIFEX_TOKEN:
            logger.warning(
                'Credentials for Transifex not found. Retrieving localized '
                'content will not be available; instead, source strings will'
                ' be displayed.'
            )
            fetch_translations = False
        else:
            fetch_translations = any(
                [
                    # Start when forced
                    (os.getenv('FORCE_TRANSLATIONS_SYNC', False) == 'true'),
                    # Start for local development
                    _segments_match(['manage.py', 'runserver'], sys.argv),
                    # Start for gunicorn
                    _segments_match(['gunicorn'], sys.argv),
                    # Start for celery worker
                    _segments_match(['celery', 'worker'], sys.argv)
                ]
            )

        # Convert from [(<lang_code>, <name>), ...]
        # to [<locale>, ...]
        # e.g. from [('en-us', 'English (USA)'), ('fr-fr', 'French (France)')]
        # to ['en_US', 'fr_FR']
        languages = [to_locale(item[0]) for item in native_settings.LANGUAGES]

        # Create the missing policy lazily to avoid import issues
        # in Django settings files
        missing_policy = parse_rendering_policy(
            native_settings.TRANSIFEX_MISSING_POLICY
        )
        error_policy = parse_error_policy(
            native_settings.TRANSIFEX_ERROR_POLICY
        )
        init(
            native_settings.TRANSIFEX_TOKEN,
            languages,
            secret=native_settings.TRANSIFEX_SECRET,
            cds_host=native_settings.TRANSIFEX_CDS_HOST,
            missing_policy=missing_policy,
            error_policy=error_policy
        )

        if fetch_translations:
            logger.info(
                'Fetching translations for languages: {}'.format(
                    ', '.join(languages)
                )
            )
            tx.fetch_translations()
            logger.info('Starting daemon for OTA translations update')
            daemon.start_daemon()
            request_finished.connect(daemon.is_daemon_running)
        else:
            logger.info(
                'Starting up without fetching translations or OTA updates'
            )



from __future__ import absolute_import, unicode_literals

import transifex.native.consts as consts
from django.template import Library, Node, TemplateSyntaxError, Variable
from django.template.base import TOKEN_BLOCK, Variable
from django.template.defaulttags import token_kwargs
from django.utils.safestring import SafeData, mark_safe
from transifex.native.django import t
from transifex.native.parsing import SourceString

register = Library()

# These keys have special meaning and cannot be used as placeholders
# inside the source strings

TRANSLATE_TAG = 't'
TRANSLATE_TAG_END = 'endt'
TRANSLATE_TAG_UNESCAPED = 'ut'
TRANSLATE_TAG_UNESCAPED_END = 'endut'


def parse_translatable_tag(parser, token):
    """Parse the given `t`/`ut` token and return its information.

    If the token is not of {% t %} type, it returns (None, None).

    See `do_translate()` for more information about supported string structure.

    :param django.template.base.Parser parser: the parser object that holds
        information on the whole template document
    :param django.template.base.Token token: holds information on a
        specific line in the template
    :param str tag: the name of the tag, i.e. 't' or 'ut'
    :return: a node object with the necessary parameters for later rendering
        a specific part of the template document, or None if the given token
        is not a `t` tag
    :return: a tuple (SourceString, dict) with the full source string information
        and the necessary parameters for rendering it, or (None, None) if the given
        token is not a `t`/`ut` tag
    :rtype: tuple
    """
    # Split the given token into its parts (includes the template tag name)
    bits = token.split_contents()

    tag_name = bits[0]
    if tag_name not in (TRANSLATE_TAG, TRANSLATE_TAG_UNESCAPED):
        return None, None

    # See if this is a multi-line or a single-line block
    # Retrieve the next lines in the template until another block is found
    # or the document ends
    next_token = None
    tokens_processed = 0
    while parser.tokens:
        if len(parser.tokens) == tokens_processed:
            break
        next_token = parser.tokens[tokens_processed]
        tokens_processed += 1
        if next_token.token_type == TOKEN_BLOCK:
            break

    is_block = False
    string = None
    params = None

    end_tag_name = (
        TRANSLATE_TAG_END if tag_name == TRANSLATE_TAG
        else TRANSLATE_TAG_UNESCAPED_END
    )
    # If `next_token` is defined, it will be of type BLOCK
    # It could be an `endt`/`endut` or any other block
    if next_token:
        # Found an `endt`/`endut` block
        if next_token.contents.strip() == end_tag_name:
            # Get the multi-line string and use it as the source string
            # Note: parser.next_token() advances the cursor in the parser
            # and shifts one item from parser.tokens
            lines = []
            for _ in range(tokens_processed):
                # Content currently looks like "\nfirst_line\nsecond_line",
                # so remove the newline char from start and end
                lines.append(parser.next_token().contents.strip('\n'))
            lines = lines[:-1]  # remove the `endt`/`endut` text
            string = '\n'.join(lines)
            params = bits[1:]
            is_block = True

    # No `endt`/`endut` block was found, so this is a one-line tag
    if not is_block:
        if len(bits) < 2:
            raise TemplateSyntaxError(
                "One-line `{start}` blocks must include a string as their first "
                "parameter, whereas multi-line `{start}` blocks should end with their "
                "respective `{end}` tags".format(
                    start=tag_name, end=end_tag_name)
            )
        string = bits[1][1:-1]  # Remove the enclosing quotes
        params = bits[2:]

    # `support_legacy=False` means we only accept "key=value" format,
    # not "<value> as <key>"
    # Provide a copy of the `params` list, because it will be consumed internally
    values_by_key = token_kwargs(list(params), parser, support_legacy=False)

    string_context = values_by_key.get('_context')
    if string_context and isinstance(string_context.var, Variable):
        raise TemplateSyntaxError(
            "Context in `{start}` blocks can only be a string, not a variable, "
            "string='{string}...', tag parameters='{params}'".format(
                start=tag_name, string=string[:40], params=' '.join(params),
            )
        )
    string_context = string_context.var if string_context else None

    # Separate the reserved keys from the user-defined ones
    extra_context = {}
    meta = {}
    for key, value in values_by_key.items():
        if key in consts.ALL_KEYS:
            if key != consts.KEY_CONTEXT:  # context is already in a variable on its own
                # We don't need the overall context for resolving,
                # we just want string and int values to get unpacked
                # without any dynamic object fetching
                # Note: the `context` argument here is Django's
                #       and has nothing to do with Transifex Native `_context`
                meta[key] = value.resolve(context=None)
        else:
            extra_context[key] = value

    return SourceString(string, string_context, **meta), extra_context


@register.tag("t")
def do_translate(parser, token):
    """Translate a one-line string or block of text using TxNative.

    The source string supports ICU Message Format. The result is HTML-escaped.

    Usage::

    One-line string:
    {% t Hello %}
    {% t "Hello there" %}
    {% t "{cnt, plural, one {One file.} other {{cnt} files.}}" cnt=files|length %}

    Block:
    {% t total_guests=guests|length host="Jenny" guest="Jo" _context="sample" %}
      {total_guests, plural, offset:1
          =0 {{host} does not give a party.}
          =1 {{host} invites {guest} to her party.}
          =2 {{host} invites {guest} and one other person to her party.}
          other {{host} invites {guest} and # other people to her party.}
        }
      }
    {% endt %}

    where `_context` is used for providing additional context to the source string.

    :param django.template.base.Parser parser: the parser object that holds
        information on the whole template document
    :param django.template.base.Token token: holds information on a
        specific line in the template
    :return: a node object with the necessary parameters for later rendering
        a specific part of the template document, or None if the given token
        is not a `t` tag
    :rtype: TranslateNode
    """
    string_obj, extra_context = parse_translatable_tag(parser, token)
    return TranslateNode(
        string=string_obj.string,
        _context=string_obj.context,
        escape=True,
        extra_context=extra_context,
    )


@register.tag("ut")
def do_translate_unescaped(parser, token):
    """Translate a one-line string or block of text using TxNative, unescaped.

    The source string support ICU Message Format. The result is NOT HTML-escaped.

    See do_translate() for usage.

    :param django.template.base.Parser parser: the parser object that holds
        information on the whole template document
    :param django.template.base.Token token: holds information on a
        specific line in the template
    :return: a node object with the necessary parameters for later rendering
        a specific part of the template document, or None if the given token
        is not a `t` tag
    :rtype: TranslateNode
    """
    string_obj, extra_context = parse_translatable_tag(parser, token)
    return TranslateNode(
        string=string_obj.string,
        _context=string_obj.context,
        escape=False,
        extra_context=extra_context,
    )


class TranslateNode(Node):
    """A template node that uses TxNative to translate a source string to
    the current language."""

    def __init__(self, string, _context=None, escape=True, extra_context=None):
        """Constructor.

        :param unicode string: the source string
        :param unicode _context: an optional context that accompanies
            the source string
        :param bool escape: if True, the returned string will be HTML-escaped,
            otherwise it won't
        :param dict extra_context: additional parameters like values
            for variable placeholders
        """
        self.string = string
        self.context = _context
        self.extra_context = extra_context
        self.escape = escape

    def render(self, context):
        """Render the template tag into a string.

        :param django.template.context.RequestContext context: the full context
            to use for applying variables, filters etc in rendering
        :return: the final translation
        :rtype: unicode
        """
        # Resolve localization context e.g. key=value vars
        localization_context = {}
        for var, val in self.extra_context.items():
            localization_context[var] = val.resolve(context)
        context.update(localization_context)

        return t(
            self.string, _context=self.context, escape=self.escape, **localization_context
        )



# -*- coding: utf-8 -*-
import mock
from django.core.management import call_command
from transifex.native.django.management.commands.pushtransifex import Command
from transifex.native.django.management.common import TranslatableFile
from transifex.native.parsing import SourceString

PYTHON_TEMPLATE = u"""
# -*- coding: utf-8 -*-

{_import}

{call1}(u'{string1}', u'désign1,désign2', param1='1', param2=2, param3=True)
{call2}(
    u'{string2}', u'opération', _comment='comment', _tags='t1,t2', _charlimit=33,
)
"""

HTML_TEMPLATE = u"""
{% load transifex %}

{content}
"""

PATH_FIND_FILES = 'transifex.native.django.management.commands' \
                  '.pushtransifex.Command._find_files'
PATH_READ_FILE = 'transifex.native.django.management.commands' \
                 '.pushtransifex.Command._read_file'
PATH_PUSH_STRINGS = 'transifex.native.django.management.commands' \
                    '.pushtransifex.Command.push_strings'


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_python_parsing_success(mock_find_files, mock_read, mock_push_strings):
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]
    mock_read.side_effect = [
        # 1.py
        PYTHON_TEMPLATE.format(
            _import='import transifex.native',
            call1='native.translate',
            call2='native.translate',
            string1=u'Le canapé',
            string2=u'Les données',
        ),
        # 2.py
        PYTHON_TEMPLATE.format(
            _import='import transifex.native as _n',
            call1='_n.translate',
            call2='_n.translate',
            string1=u'Le canapé 2',
            string2=u'Les données 2',
        ),
        # 3.py
        PYTHON_TEMPLATE.format(
            _import='from transifex.native import translate',
            call1='translate',
            call2='translate',
            string1=u'Le canapé 3',
            string2=u'Les données 3',
        ),
    ]

    expected = [
        # 1.py
        SourceString(u'Le canapé', u'désign1,désign2'),
        SourceString(
            u'Les données', u'opération', _comment='comment', _tags='t1,t2',
            _charlimit=33,
        ),
        # 2.py
        SourceString(u'Le canapé 2', u'désign1,désign2'),
        SourceString(
            u'Les données 2', u'opération', _comment='comment', _tags='t1,t2',
            _charlimit=33,
        ),
        # 3.py
        SourceString(u'Le canapé 3', u'désign1,désign2'),
        SourceString(
            u'Les données 3', u'opération', _comment='comment', _tags='t1,t2',
            _charlimit=33,
        ),
    ]
    compare(expected)


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_template_parsing(mock_find_files, mock_read, mock_push_strings):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        # 1.html
        HTML_TEMPLATE.replace(
            '{content}',
            u'<p>{% t "<b>Strong</b> {a}" a="A" _context="c1,c2" '
            u'_tags="t1,t2" _comment="comment1" _charlimit=22 %}</p>\n'

            u'<p>{% ut "παράδειγμα {b}" b="B" _context="c1,c2" '
            u'_tags="t1,t2" _comment="comment2" _charlimit=33 %}</p>'
        ),
        # 1.txt
        HTML_TEMPLATE.replace(
            '{content}',
            u'{% t _context="c1,c2" _tags="t1,t2" _comment="co1" _charlimit=22 %}\n'
            u'This is a short string\n'
            u'{% endt %}\n'

            u'{% t _context="c1,c2" _tags="t1,t2" _comment="co2" _charlimit=33 %}\n'
            u'This is not a shorter string\n'
            u'{% endt %}'
        ),
    ]

    expected = [
        # 1.html
        SourceString(
            u'<b>Strong</b> {a}', 'c1,c2', _tags='t1,t2', _comment="comment1",
            _charlimit=22,
        ),
        SourceString(
            u'παράδειγμα {b}', 'c1,c2', _tags='t1,t2', _comment="comment2",
            _charlimit=33,
        ),

        # 1.txt
        SourceString(
            u'This is a short string', 'c1,c2', _tags='t1,t2', _comment="co1",
            _charlimit=22,
        ),
        SourceString(
            u'This is not a shorter string', 'c1,c2', _tags='t1,t2', _comment="co2",
            _charlimit=33,
        ),
    ]
    compare(expected)


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_no_detection_for_non_transifex(mock_find_files, mock_read, mock_push_strings):
    """No strings should be detected if a format other than Transifex Native
    is used in Python files and templates.
    """
    mock_find_files.return_value = [
        # 2 files with valid extension but no translatable content
        TranslatableFile('dir4/dir5', 'empty.py', 'locdir1'),
        TranslatableFile('dir4/dir5', 'empty.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        # empty.py - shouldn't detect any strings as non-transifex
        PYTHON_TEMPLATE.format(
            _import='from django.utils.translation import ugettext_lazy as _',
            call1='_',
            call2='_',
            string1=u'A Django string',
            string2=u'Another Django string',
        ),
        # empty.txt - shouldn't detect any strings as non-transifex
        (
            u'{% load i18n %}\n'
            u'{% trans "A Django string %}\n'
            u'{% blocktrans %}Another Django string{% endblocktrans %}',
        ),
    ]
    compare(expected=[])


def compare(expected):
    """Run the command and compare the detected strings with the expected ones.

    :param list expected: a list of SourceString objects
    """
    command = Command()
    call_command(command)
    # command.string_collection.strings is like: {<key>: <SourceString>}
    found = command.string_collection.strings.values()
    assert set(found) == set(expected)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import sys

from transifex.common.console import Color, prompt
# The reviewer has decided to reject the changes on a specific string
from transifex.native.tools.migrations.models import Confidence

# The reviewer has decided to reject the changes on a specific string or file
REVIEW_REJECT = 1

# The reviewer has decided to reject the changes of all strings or all files
REVIEW_REJECT_ALL = 2

# The reviewer has decided to accept the change of one string or one file
REVIEW_ACCEPT = 3

# The reviewer has decided to accept the changes of all strings or all files
REVIEW_ACCEPT_ALL = 4

# The reviewer has decided to mark a specific string for later review
REVIEW_MARK_STRING = 5

# The reviewer has decided to mark a specific file for later review
REVIEW_MARK_FILE = 6

# When the user is prompted to review a change (file or string)
# these are the valid responses
ACCEPT_CHOICE = 'A'  # accept the change
ACCEPT_ALL_CHOICE = 'AA'  # accept all changes
REJECT_CHOICE = 'R'  # reject the change
REJECT_ALL_CHOICE = 'RR'  # reject all change
MARK_CHOICE = 'M'  # mark for later review (inside the file, if saved)
PRINT_CHAR = 'P'  # print the diff

MARK_REVIEW_STRING = 'Transifex Native: REVIEW_STRING'
MARK_REVIEW_FILE = 'Transifex Native: REVIEW_FILE'


def prompt_to_start(total_files):
    """Prompt the user before starting the migration.

    If the user chooses to not go through with it, sys.exit() is called.

    :param int total_files: the total number of files to migrate
    """
    Color.echo(
        '\nFound {} file(s) to check for translatable strings.'.format(
            total_files)
    )
    while True:
        reply = prompt(
            Color.format(
                '[opt](Y)[end] Yes [opt](N)[end] No'
            ),
            description='Are you sure you want to continue?',
            default='N',
        )
        reply = reply.upper()
        if reply == 'Y':
            return
        elif reply == 'N':
            Color.echo('\n[high]Migration aborted.[end]')
            sys.exit(1)


class ReviewPolicy(object):
    """Determines if and how each migration (string or file) will be reviewed
    by the user, before being accepted for save.

    It is meant to be subclassed, in order to provide custom functionality.
    """

    def __init__(self):
        # This determines review comments will appear
        # inside the migrated file. It is dynamic because different files
        # support different comment formats, e.g.
        # - '# Comment' in Python
        # - '<!-- Comment -->' in HTML
        self._comment_format = '{}'

    def set_comment_format(self, comment_format):
        """Define the comment format to use when a review policy adds
        comments to the reviewed file.

        Example:
        >>> set_comment_format('# {}')

        :param unicode comment_format: the format to use; must include '{}'
        :raise ValueError: if `comment_format` does not include '{}'
        """
        if '{}' not in comment_format:
            raise ValueError('The provided `comment_format` must include {}')
        self._comment_format = comment_format

    def should_review_strings(self):
        """Returns whether or not a policy object wants to review individual
        strings.

        Although policy subclasses can only review *some* strings, and not all,
        this method is important for optimization, so that the objects that
        use a policy object know if they need to feed each string migration
        to the review policy or not.
        """
        return False

    def review_string(self, string_migration, string_cnt, strings_total):
        """Optionally give the chance to the user to review a string migration
        before accepting it.

        By default this class automatically accepts the change
        (i.e. returns an "accept" directive), but subclasses can provide
        a custom implementation.

        :param StringMigration string_migration: the migration of a single string
        :param int string_cnt: the current index of the migrated string (0-based)
        :param int strings_total: the total number of strings in the corresponding
            file migration
        :return: an integer directive that determines what to do with the string
        :rtype: int
        """
        return REVIEW_ACCEPT

    def review_file(self, file_migration):
        """Optionally give the chance to the user to review a file migration
        before accepting it.

        :param FileMigration file_migration: the migration of a whole file
        :return: an integer directive that determines what to do with the string
        :rtype: int
        """
        return REVIEW_ACCEPT

    def prompt_for_string(self, string_migration, string_cnt, strings_total):
        """Prompt the user to review the string migration and decide what to do.

        :param StringMigration string_migration: the migration object
        :param int string_cnt: the current index of the migrated string (0-based)
        :param int strings_total: the total number of strings in the corresponding
            file migration
        :return: an integer directive that determines what to do with the string
        :rtype: int
        """
        Color.echo(
            '\n[{cnt}/{total}] [prompt]'
            'Please review the following change:[end] '
            '(confidence={confidence})'.format(
                cnt=string_cnt + 1,
                total=strings_total,
                confidence=(
                    '{}' if string_migration.confidence == Confidence.HIGH
                    else '[warn]{}[end]'
                ).format(
                    Confidence.to_string(string_migration.confidence)
                ),
            )
        )
        Color.echo('[red]{}[end]'.format(string_migration.original))
        Color.echo('[green]{}[end]'.format(string_migration.new))
        while True:
            reply = prompt(
                Color.format(
                    '[opt](A)[end] Accept '
                    '[opt](R)[end] Reject '
                    '[opt](M)[end] Mark for later review '
                    '\n[opt](AA)[end] Accept all remaining '
                    '[opt](RR)[end] Reject all remaining '
                ),
                default=str(ACCEPT_CHOICE),
            )
            reply = reply.upper()
            if reply == ACCEPT_CHOICE:
                Color.echo('✅️ Change accepted')
                return REVIEW_ACCEPT

            elif reply == ACCEPT_ALL_CHOICE:
                Color.echo('✅ ✅️ All remaining changes accepted')
                return REVIEW_ACCEPT_ALL

            elif reply == REJECT_CHOICE:
                string_migration.revert()
                Color.echo('❌ Change rejected')
                return REVIEW_REJECT

            elif reply == REJECT_ALL_CHOICE:
                string_migration.revert()
                Color.echo('❌ ❌ All remaining changes rejected')
                return REVIEW_REJECT_ALL

            elif reply == MARK_CHOICE:
                string_migration.update(
                    '', self._comment_format.format(MARK_REVIEW_STRING),
                    append=False,
                )
                Color.echo('📝 Change marked for later review')
                return REVIEW_MARK_STRING

    def prompt_for_file(self, file_migration):
        """Prompt the user to review the file migration and decide what to do.

        :param FileMigration file_migration: the migration object
        :return: an integer directive that determines what to do with the file
        :rtype: int
        """
        modified_strings = file_migration.modified_strings
        while True:
            reply = prompt(
                Color.format(
                    '[opt](A)[end] Accept '
                    '[opt](R)[end] Reject '
                    '[opt](M)[end] Mark for later review '
                    '[opt](P)[end] Print file diff '
                    '\n[opt](RR)[end] Reject all remaining '
                    '[opt](AA)[end] Accept all remaining'
                ),
                default=str(ACCEPT_CHOICE),
            )
            reply = reply.upper()
            if reply == ACCEPT_CHOICE:
                Color.echo('✅️ Changes accepted')
                return REVIEW_ACCEPT

            elif reply == ACCEPT_ALL_CHOICE:
                Color.echo('\n[warn]WARNING![end]')
                reply = self._yes_no(
                    'If you continue, all changes in the remaining files '
                    'will be accepted. Are you sure you want to continue?',
                    no_message='Aborted.'
                )
                if reply is True:
                    Color.echo('✅ ✅ Changes in all remaining files accepted')
                    return REVIEW_ACCEPT_ALL

            elif reply == REJECT_CHOICE:
                file_migration.revert()
                Color.echo('❌ Changes in file rejected')
                return REVIEW_REJECT

            elif reply == REJECT_ALL_CHOICE:
                Color.echo('\n[warn]WARNING![end]')
                reply = self._yes_no(
                    'If you continue, all changes in the remaining files '
                    'will be rejected. Are you sure you want to continue?',
                    no_message='Aborted.'
                )
                if reply is True:
                    Color.echo('❌ ❌ Changes in all remaining files rejected')
                    return REVIEW_REJECT_ALL

            elif reply == MARK_CHOICE:
                file_migration.strings[0].update(
                    '', self._comment_format.format(MARK_REVIEW_FILE),
                    append=False,
                )
                Color.echo('📝 File marked for later review')
                return REVIEW_MARK_STRING

            elif reply == PRINT_CHAR:
                Color.echo(
                    '[prompt]These are the modified strings[end]'
                    ' (the rest of the file is omitted)'
                )
                for string_migration in modified_strings:
                    Color.echo('[red]{}[end]'.format(
                        string_migration.original))
                    Color.echo('[green]{}[end]'.format(string_migration.new))
                    print('')

    def _yes_no(self, description, yes_message=None, no_message=None):
        """Prompts the user to reply to a Yes/No question.

        :param basestr description: the message to display before prompting
        :param basestr yes_message: the message to display if user accepts
        :param basestr no_message: the message to display is user declines
        :return: True if the user chose to go through, false otherwise
        :rtype: bool
        """
        while True:
            reply = prompt(
                Color.format('[opt](Y)[end] Yes [opt](N)[end] No'),
                description=description,
                default='N',
            )
            reply = reply.upper()
            if reply == 'Y':
                if yes_message:
                    Color.echo('[high]{}[end]'.format(yes_message))
                return True
            elif reply == 'N':
                if no_message:
                    Color.echo('[high]{}[end]'.format(no_message))
                return False


class NoopReviewPolicy(ReviewPolicy):
    """Never prompts the user for anything."""
    pass


class FileReviewPolicy(ReviewPolicy):
    """Prompts the user to review each file."""

    def review_file(self, file_migration):
        return self.prompt_for_file(file_migration)


class LowConfidenceFileReviewPolicy(ReviewPolicy):
    """Prompts the user to review each file that includes at least one string
    with low confidence."""

    def review_file(self, file_migration):
        total_low_confidence = len(
            [
                string for string in file_migration.modified_strings
                if string.confidence == Confidence.LOW
            ]
        )
        if total_low_confidence > 0:
            return self.prompt_for_file(file_migration)

        return REVIEW_ACCEPT


class StringReviewPolicy(ReviewPolicy):
    """Prompts the user to review each string."""

    def should_review_strings(self):
        return True

    def review_string(self, string_migration, string_cnt, strings_total):
        return self.prompt_for_string(string_migration, string_cnt, strings_total)


class LowConfidenceStringReviewPolicy(ReviewPolicy):
    """Prompts the user to review each string that has low confidence."""

    def should_review_strings(self):
        return True

    def review_string(self, string_migration, string_cnt, strings_total):
        if string_migration.confidence == Confidence.LOW:
            return self.prompt_for_string(string_migration, string_cnt, strings_total)



from __future__ import unicode_literals

import fnmatch
import io
import os

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import BaseCommand
from django.core.management.utils import handle_extensions
from django.utils.functional import cached_property
from transifex.common.console import Color
from transifex.native.django.management.common import TranslatableFile
from transifex.native.django.tools.migrations.templatetags import \
    DjangoTagMigrationBuilder
from transifex.native.tools.migrations.models import Confidence
from transifex.native.tools.migrations.review import (
    REVIEW_ACCEPT_ALL, REVIEW_REJECT_ALL, FileReviewPolicy,
    LowConfidenceFileReviewPolicy, LowConfidenceStringReviewPolicy,
    NoopReviewPolicy, StringReviewPolicy, prompt_to_start)
from transifex.native.tools.migrations.save import (BackupSavePolicy,
                                                    InPlaceSavePolicy,
                                                    NewFileSavePolicy,
                                                    NoopSavePolicy)

SAVE_POLICY_OPTIONS = {
    'dry-run': 'no changes will be saved\n',
    'new': 'migrated content will be saved in a new file, '
           'named <filename>__native.<extension>\n',
    'backup': 'migrated content will be saved directly in the original '
              'file path, and a backup will also be saved '
              'in <filename>.<extension>.bak\n',
    'replace': 'migrated content will be saved in the original file',
}

REVIEW_POLICY_OPTIONS = {
    'none': 'everything will be done automatically without '
            'having a chance to review anything\n',
    'file': 'you get a chance to review each migrated file '
            'before it is saved\n',
    'string': 'you get a chance to review each string of each file '
              'before the file is saved\n',
    'file-low': 'you get a chance to review each migrated file '
                'that includes at least one string that has '
                'a low migration confidence\n',
    'string-low': 'you get a chance to review each string '
                  'that has a low migration confidence\n',
}


def pretty_options(options_dict):
    items = [(k, v) for k, v in options_dict.items()]
    return '\n'.join([' - "{}": {}'.format(x[0], x[1]) for x in items])


class Command(BaseCommand):
    """Migrates files using the Django i18n syntax to Transifex Native syntax."""

    def add_arguments(self, parser):
        parser.add_argument(
            '--domain', '-d', default='django', dest='domain',
            help='The domain of the message files (default: "django").',
        )
        parser.add_argument(
            '--extension', '-e', dest='extensions', action='append',
            help='The file extension(s) to examine (default: "html,txt,py", or "js" '
                 'if the domain is "djangojs"). Separate multiple extensions with '
                 'commas, or use -e multiple times.',
        )
        parser.add_argument(
            '--file', '-f', dest='files', action='append',
            help='The relative paths of the files to migrate. Separate multiple paths '
                 'with commas, or use -f multiple times.',
        )
        parser.add_argument(
            '--path', '-p', dest='path',
            help='The path of the files to migrate. Finds files recursively.',
        )
        parser.add_argument(
            '--save', dest='save_policy', default='new',
            help=(
                'Determines where the migrated content will be saved: \n' +
                pretty_options(SAVE_POLICY_OPTIONS)
            ),
        )
        parser.add_argument(
            '--review', dest='review_policy', default='file',
            help=(
                'Determines where the migrated content will be saved: \n' +
                pretty_options(REVIEW_POLICY_OPTIONS)
            ),
        )

    def handle(self, *args, **options):
        self.ignore_patterns = []
        self.verbosity = options['verbosity']
        self.path = options['path']
        self.files = options['files']

        self.save_policy = self._create_save_policy(options['save_policy'])
        self.review_policy = self._create_review_policy(
            options['review_policy'])

        self.extensions = handle_extensions(['html', 'txt', 'py'])
        self.stats = {
            'processed_files': 0, 'migrations': [], 'saved': [], 'errors': [],
        }

        # Show an intro message
        self._show_intro(options)

        # If specific files are defined, use those
        if self.files:
            dirpath = os.getcwd()
            files = [
                TranslatableFile(dirpath, filename)
                for filename in self.files
            ]
        # Else search the file path for supported files
        else:
            files = self._find_files(self.path)

        # Execute the migration
        self.django_migration_builder = DjangoTagMigrationBuilder()
        self.migrate_files(files)

    def _show_intro(self, options):
        """Show an introductory message to help the user understand what is going on.

        :param dict options: the configuration options of the command
        """
        Color.echo(
            '[high]'
            '\n####################################################################\n'
            'Running migration from Django i18n syntax to Transifex Native syntax\n'
            '[end]'
            'This migration is idempotent, so its output should not change if run\n'
            'multiple times.'
        )
        Color.echo('\n[high]Configuration:[end]')
        Color.echo('- [opt]Path:[end] [file]{}[end]'.format(self.path))
        Color.echo(
            '- [opt]Save policy:[end] [high]{}[end] -> {}'.format(
                options['save_policy'],
                SAVE_POLICY_OPTIONS[options['save_policy']],
            ).strip()
        )
        Color.echo(
            '- [opt]Review policy:[end] [high]{}[end] -> {}'.format(
                options['review_policy'],
                REVIEW_POLICY_OPTIONS[options['review_policy']],
            ).strip()
        )

    def migrate_files(self, files):
        """Search all related files, detect Django i18n translate hooks and migrate
        them to Transifex syntax.

        :param list files: a list of TranslatableFile objects
        """
        files_total = len(files)
        prompt_to_start(files_total)

        # Loop through each file, migrate, ask for user review if applicable,
        # save to disk if applicable
        accept_remaining_files = False
        for file_cnt, translatable_file in enumerate(files):
            _, extension = os.path.splitext(translatable_file.file)
            comment_format = '# {}\n' if extension == '.py' else '<!-- {} -->\n'
            self.review_policy.set_comment_format(comment_format)

            Color.echo(
                '\n---- [{cnt}/{total}] Migrating [file]{path}[end]...'.format(
                    path=translatable_file.path,
                    cnt=file_cnt + 1,
                    total=files_total,
                )
            )
            file_migration = self._migrate_file(translatable_file)
            if not file_migration:
                self.stats['processed_files'] += 1
                continue

            modified_strings = file_migration.modified_strings
            total_modified = len(modified_strings)
            total_low_confidence = len(
                [x for x in modified_strings if x.confidence == Confidence.LOW]
            )
            Color.echo(
                '[prompt]{strings} strings were modified '
                ' ([warn]{low}[end] with LOW confidence)[end]'.format(
                    strings=total_modified,
                    low=total_low_confidence,
                )
            )
            if not total_modified:
                continue

            # If the review policy says so, prompt the user for each string
            # If this returns True, it doesn't necessarily mean that the user
            # will be prompted, as the actual review policy may use
            # additional filters, e.g. only prompt for strings with low
            # confidence
            if self.review_policy.should_review_strings():
                reject_remaining = False
                for string_index in range(total_modified):
                    string_migration = modified_strings[string_index]

                    # The rest of the string test_migrations should be reverted
                    # based on the user's choice
                    if reject_remaining:
                        string_migration.revert()

                    # Optionally prompt the user to review the migration
                    else:
                        # Give the user the option to review
                        # May modify `string_migration` in place
                        result = self.review_policy.review_string(
                            string_migration, string_index, total_modified
                        )
                        # The user has chosen to accept the changes
                        # in all remaining strings. Break so that
                        # there will be no more prompts for the rest
                        # of the strings
                        if result == REVIEW_ACCEPT_ALL:
                            break

                        # The user has chosen to reject the changes in all
                        # remaining strings. Set the flag to True, so that
                        # it will revert all changes for the rest strings
                        # in the loop
                        if result == REVIEW_REJECT_ALL:
                            reject_remaining = True

            # If the review policy says so, prompt the user for each file
            if accept_remaining_files is False:
                result = self.review_policy.review_file(file_migration)

                # The user has chosen to reject all remaining files
                # Break, so that we exit the outer (file) loop
                if result == REVIEW_REJECT_ALL:
                    break

                # The user has chosen to accept all remaining files
                # Set the flag, so that the file review policy won't be used
                # for the remaining file test_migrations
                if result == REVIEW_ACCEPT_ALL:
                    accept_remaining_files = True

            # If the save policy says so, save the changes
            if file_migration.modified_strings:
                saved, error_type = self.save_policy.save_file(file_migration)
            else:
                saved, error_type = False, None

            # Update stats
            self.stats['processed_files'] += 1
            self.stats['migrations'].append(
                (translatable_file.path, file_migration)
            )
            if saved:
                self.stats['saved'].append(file_migration)
            elif error_type is not None:
                self.stats['errors'].append(file_migration)

        self._show_results(files)

    def _migrate_file(self, translatable_file):
        """Extract source strings from the given file.

        Supports both Python files and Django template files.

        :param TranslatableFile translatable_file: the file to search
        :return: an object with the migration info
        :rtype: FileMigration
        """
        self.verbose('Processing file %s in %s' % (
            translatable_file.file, translatable_file.dirpath
        ))
        encoding = (
            settings.FILE_CHARSET if self.settings_available
            else 'utf-8'
        )
        try:
            src_data = self._read_file(translatable_file.path, encoding)
        except UnicodeDecodeError as e:
            self.verbose(
                'UnicodeDecodeError: skipped file %s in %s (reason: %s)' % (
                    translatable_file.file, translatable_file.dirpath, e,
                )
            )
            return None

        _, extension = os.path.splitext(translatable_file.file)

        # Python file
        if extension == '.py':
            return None  # TODO

        # Template file
        return self.django_migration_builder.build_migration(
            src_data, translatable_file.path, encoding,
        )

    def _show_results(self, files):
        """Show a detailed report of how the migration went.

        :param list files: a list of TranslatableFile objects
        """
        Color.echo('\n\n[high]Migration completed![end]')
        Color.echo('--------------------')
        Color.echo('[high]Total files found:[end] [warn]{}[end]'.format(
            len(files))
        )
        Color.echo('[high]Total files processed:[end] [warn]{}[end]'.format(
            self.stats['processed_files'])
        )
        modified = [
            file_migration for _, file_migration in self.stats['migrations']
            if len(file_migration.modified_strings)
        ]
        Color.echo('[high]Total migrations created:[end] [warn]{}[end]'.format(
            len(modified)
        ))
        Color.echo('[high]Total files saved:[end] [warn]{}[end]'.format(
            len(self.stats['saved'])
        ))
        Color.echo(
            '\n'.join([
                '- [file]{}[end]'.format(x.filename)
                for x in self.stats['saved']
            ])
        )
        Color.echo('[high]Total errors:[end] [warn]{}[end]'.format(
            len(self.stats['errors'])
        ))
        Color.echo(
            '\n'.join([
                '- [warn]{}[end]'.format(x.filename)
                for x in self.stats['errors']
            ])
        )
        Color.echo('')

    def _find_files(self, root):
        """Get all files in the given root.

        :param basestring root: the root path to search in
        :return: a list of TranslatableFile objects
        :rtype: list
        """

        def is_ignored(path, ignore_patterns):
            """Check if the given path should be ignored or not."""
            filename = os.path.basename(path)

            def ignore(pattern):
                return fnmatch.fnmatchcase(filename, pattern) or fnmatch.fnmatchcase(
                    path, pattern)

            return any(ignore(pattern) for pattern in ignore_patterns)

        ignore_patterns = [os.path.normcase(p) for p in self.ignore_patterns]
        dir_suffixes = {'%s*' % path_sep for path_sep in {'/', os.sep}}
        norm_patterns = []
        for p in ignore_patterns:
            for dir_suffix in dir_suffixes:
                if p.endswith(dir_suffix):
                    norm_patterns.append(p[:-len(dir_suffix)])
                    break
            else:
                norm_patterns.append(p)

        all_files = []
        ignored_roots = []
        if self.settings_available:
            ignored_roots = [os.path.normpath(p) for p in
                             (settings.MEDIA_ROOT, settings.STATIC_ROOT) if p]

        for dirpath, dirnames, filenames in os.walk(root, topdown=True,
                                                    followlinks=False):
            for dirname in dirnames[:]:
                if (is_ignored(os.path.normpath(os.path.join(dirpath, dirname)),
                               norm_patterns) or
                        os.path.join(os.path.abspath(dirpath),
                                     dirname) in ignored_roots):
                    dirnames.remove(dirname)
                    self.verbose('Ignoring directory %s' % dirname)
                elif dirname == 'locale':
                    dirnames.remove(dirname)

            for filename in filenames:
                file_path = os.path.normpath(os.path.join(dirpath, filename))
                file_ext = os.path.splitext(filename)[1]
                if file_ext not in self.extensions or is_ignored(file_path,
                                                                 self.ignore_patterns):
                    self.verbose('Ignoring file %s in %s' %
                                 (filename, dirpath))
                else:
                    all_files.append(
                        TranslatableFile(dirpath.lstrip('./'), filename)
                    )
        return sorted(all_files)

    def _read_file(self, path, encoding):
        with io.open(path, 'r', encoding=encoding) as fp:
            return fp.read()

    @cached_property
    def settings_available(self):
        try:
            settings.LOCALE_PATHS
        except ImproperlyConfigured:
            self.verbose("Running without configured settings.")
            return False
        return True

    def verbose(self, msg):
        if self.verbosity > 1:
            print(msg)

    def output(self, msg):
        print(msg)

    def _create_save_policy(self, policy_id):
        """Create the save policy object that corresponds to the given ID.

        :param str policy_id: the ID of the policy to create, as expected
            in the command parameters
        :return: a SavePolicy subclass
        :rtype: SavePolicy
        """
        policy_id = policy_id.lower()
        if policy_id == 'dry-run':
            return NoopSavePolicy()
        elif policy_id == 'new':
            return NewFileSavePolicy()
        elif policy_id == 'backup':
            return BackupSavePolicy()
        elif policy_id == 'replace':
            return InPlaceSavePolicy()

        raise AttributeError('Invalid save policy ID={}'.format(policy_id))

    def _create_review_policy(self, policy_id):
        """Create the review policy object that corresponds to the given ID.

        :param str policy_id: the ID of the policy to create, as expected
            in the command parameters
        :return: a ReviewPolicy subclass
        :rtype: ReviewPolicy
        """
        if policy_id == 'none':
            return NoopReviewPolicy()
        elif policy_id == 'file':
            return FileReviewPolicy()
        elif policy_id == 'string':
            return StringReviewPolicy()
        elif policy_id == 'file-low':
            return LowConfidenceFileReviewPolicy()
        elif policy_id == 'string-low':
            return LowConfidenceStringReviewPolicy()

        raise AttributeError('Invalid review policy ID={}'.format(policy_id))



from django.conf import settings
from django.utils.translation import get_language, to_locale
from transifex.native import tx


def translate(string, _context=None, escape=True, **params):
    """Translate the given source string to the current language.

    A convenience wrapper that uses the current language of a Django app.

    If there are any placeholders to replace, they need to be passed as kwargs.

    :param unicode string: the source string to get the translation for
    :param unicode _context: an optional context that gives more information about
        the source string
    :param bool escape: if True, the returned string will be HTML-escaped,
        otherwise it won't
    :return: the final translation in the current language
    :rtype: unicode
    """
    is_source = get_language() == settings.LANGUAGE_CODE
    locale = to_locale(get_language())  # e.g. from en-us to en_US
    return tx.translate(
        string,
        locale,
        _context=_context,
        is_source=is_source,
        escape=escape,
        params=params,
    )


def utranslate(string, _context=None, **params):
    """Translate the given source string to the current language, without HTML escaping.

    While the given `string` is not escaped, all `params` are, before replacing
    the placeholders inside the `string`.

    A convenience wrapper that uses the current language of a Django app.

    If there are any placeholders to replace, they need to be passed as kwargs.

    :param unicode string: the source string to get the translation for
    :param unicode _context: an optional context that gives more information about
        the source string
    :return: the final translation in the current language
    :rtype: unicode
    """
    return translate(string, _context, escape=False, **params)



# -*- coding: utf-8 -*-
from transifex.native.cache import MemoryCache


class TestMemoryCache(object):
    """Tests the functionality of the MemoryCache class."""

    def test_returns_none_if_not_exists(self):
        cache = MemoryCache()
        assert cache.get('missing_key', 'en') is None

    def test_returns_entry_if_exists(self):
        cache = MemoryCache()
        cache.update(
            {
                'en': {
                    'table': {
                        'string': u'A table'
                    },
                    'chair': {
                        'string': u'A chair'
                    },
                },
                'el': {
                    'table': {
                        'string': u'Ένα τραπέζι'
                    },
                    'chair': {
                        'string': u'Μια καρέκλα'
                    },
                },
            },
        )
        assert cache.get('table', 'en') == u'A table'
        assert cache.get('chair', 'en') == u'A chair'
        assert cache.get('table', 'el') == u'Ένα τραπέζι'
        assert cache.get('chair', 'el') == u'Μια καρέκλα'
        assert cache.get('invalid', 'en') is None
        assert cache.get('invalid', 'el') is None



import click


class Color:
    """Convenience class for adding color to console output."""

    CYAN = '\033[36m'
    HIGHLIGHT = '\033[1m'
    GREEN = '\033[32m'
    PINK = '\033[91m'
    RED = '\033[31m'
    YELLOW = '\033[33m'
    END = '\033[0m'

    @staticmethod
    def format(string):
        """Format the given string, adding color support."""
        return (
            string.replace('[high]', Color.HIGHLIGHT)
            .replace('[cyan]', Color.CYAN)
            .replace('[end]', Color.END)
            .replace('[green]', Color.GREEN)
            .replace('[red]', Color.RED)
            .replace('[opt]', Color.PINK)
            .replace('[warn]', Color.PINK)
            .replace('[prompt]', Color.YELLOW)
            .replace('[file]', Color.CYAN)
        )

    @staticmethod
    def echo(string):
        """Print to the console with color support."""
        print(Color.format(string))


def display(*messages):
    """Print all given strings respecting any color markup."""
    for msg in messages:
        Color.echo(msg)


def prompt(prompt_msg, description=None, default=None, new_line=False, vtype=None):
    """Prompt the user to enter a reply.

    :param str prompt_msg: the prompt message
    :param basestr description: an optional description to show above the prompt message
    :param basestr default: an optional default value
    :param bool new_line: if True, an empty line will be printed before the prompt
    :param type vtype: the expected type of the input
    :return: the prompt object returned by Click
    :rtype: Any
    """
    if new_line:
        Color.echo('')

    if description:
        Color.echo('[prompt]{description}[end]'.format(
            description=description))

    return click.prompt(prompt_msg, default=default, type=(vtype or str))



import logging
import threading
import time

from transifex.native import tx

logger = logging.getLogger('transifex.native.threading')


class DaemonicThread(threading.Thread):
    """A daemon thread that implements the logic of fetching
    translations periodically."""
    daemon = True
    should_exit = False

    def start_daemon(self, interval=10):
        """Start the daemon.

        Calls `threading.Thread.start()` to schedule execution in a different thread.

        :param int interval: the interval the daemon will use when fetching
            translations.
        """
        if self.is_daemon_running(log_errors=False):
            return False
        self.interval = interval
        self.start()
        return True

    def run(self, *args, **kwargs):
        """
        Fetches translations in an interval. Will not stop if exceptions are
        raised.
        """
        while not self.should_exit:
            logger.debug('Will fetch translations')
            try:
                tx.fetch_translations()
            except Exception as e:
                logger.exception(
                    'Fetching daemon exception: {}'.format(str(e)))
            time.sleep(self.interval)

    def is_daemon_running(self, log_errors=True, **kwargs):
        """Return whether the daemon is running or not.

        Checks both `is_alive` and `isAlive` (python compatibility)

        :param bool log_errors: Whether to log errors if the thread is
            not alive or not. Useful when using this method in your application
            so that you get notified if the daemon has stopped for some reason.
        """
        is_running = getattr(self, 'is_alive', self.isAlive)()
        if not is_running and log_errors:
            logger.error('Fetching daemon error: The daemon is not running!')
        return is_running

    def stop_daemon(self):
        """Set the `should_exit` variable (which should force the thread to
        stop execution), then `join`s the thread.

        Meant to be used if you explicitly want to kill the thread which is
        usually not necessary since it's a demonic thread. This function can
        potentially **block** for `interval` seconds, so use wisely.
        """

        self.should_exit = True
        self.join()


daemon = DaemonicThread()



from __future__ import unicode_literals

import fnmatch
import io
import json
import os

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import BaseCommand
from django.core.management.utils import handle_extensions
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from transifex.native import tx
from transifex.native.django.management.common import (NO_LOCALE_DIR,
                                                       SourceStringCollection,
                                                       TranslatableFile)
from transifex.native.django.utils.templates import \
    extract_transifex_template_strings
from transifex.native.parsing import Extractor


class Command(BaseCommand):
    """Detects translatable strings in Django templates and Python files,
      based on the syntax of Transifex Native and pushes them as source strings
      to Transifex."""

    def add_arguments(self, parser):
        parser.add_argument(
            '--domain', '-d', default='django', dest='domain',
            help='The domain of the message files (default: "django").',
        )
        parser.add_argument(
            '--purge', '-p', action='store_true', dest='purge', default=False,
            help='Replace the entire resource content with the pushed content '
                 'of this request. If not provided (the default), then append '
                 'the source content of this request to the existing resource '
                 'content.',
        )
        parser.add_argument(
            '--symlinks', '-s', action='store_true', dest='symlinks', default=False,
            help='Follows symlinks to directories when examining source code '
                 'and templates for translation strings.',
        )
        parser.add_argument(
            '--extension', '-e', dest='extensions', action='append',
            help='The file extension(s) to examine (default: "html,txt,py", or "js" '
                 'if the domain is "djangojs"). Separate multiple extensions with '
                 'commas, or use -e multiple times.',
        )

    def handle(self, *args, **options):
        extensions = options['extensions']
        self.ignore_patterns = []
        self.purge = options['purge']
        self.symlinks = options['symlinks']
        self.domain = options['domain']
        self.verbosity = options['verbosity']
        self.locale_paths = []
        self.default_locale_path = None

        self.string_collection = SourceStringCollection()

        if self.domain == 'djangojs':
            exts = extensions if extensions else ['js']
        else:
            exts = extensions if extensions else ['html', 'txt', 'py']
        self.extensions = handle_extensions(exts)

        # Create an extractor for Python files, to reuse for all files
        self.python_extractor = Extractor()
        # Support `t()` and `ut()` calls made on the Django module.
        self.python_extractor.register_functions(
            'transifex.native.django.t',
            'transifex.native.django.ut',
        )

        self.stats = {'processed_files': 0, 'strings': []}

        # Search all related files and collect translatable strings,
        # storing them in `self.string_collection`
        self.collect_strings()

        # Push the strings to the CDS
        self.push_strings()

    def collect_strings(self):
        """Search all related files, collect and store translatable strings.

        Stores found strings in `self.string_collection`.
        """
        self.output('Parsing all files to detect translatable content...')
        files = self._find_files('.')
        for f in files:
            extracted = self._extract_strings(f)
            self.string_collection.extend(extracted)
            self.stats['processed_files'] += 1
            if len(extracted):
                self.stats['strings'].append((f.file, len(extracted)))

        self._show_collect_results()

    def push_strings(self):
        """Push strings to the CDS."""
        total = len(self.string_collection.strings)
        if total == 0:
            self.output('There are no strings to push to Transifex.')
            return

        self.output('Pushing {} source strings to Transifex...'.format(total))
        status_code, response_content = tx.push_source_strings(
            self.string_collection.strings.values(), self.purge
        )
        self._show_push_results(status_code, response_content)

    def _extract_strings(self, translatable_file):
        """Extract source strings from the given file.

        Supports both Python files and Django template files.

        :param TranslatableFile translatable_file: the file to search
        :return: a list of SourceString objects
        :rtype: list
        """
        self.verbose('Processing file %s in %s' % (
            translatable_file.file, translatable_file.dirpath
        ))
        encoding = (
            settings.FILE_CHARSET if self.settings_available
            else 'utf-8'
        )
        try:
            src_data = self._read_file(translatable_file.path, encoding)
        except UnicodeDecodeError as e:
            self.verbose(
                'UnicodeDecodeError: skipped file %s in %s (reason: %s)' % (
                    translatable_file.file, translatable_file.dirpath, e,
                )
            )
            return None

        _, extension = os.path.splitext(translatable_file.file)

        # Python file
        if extension == '.py':
            return self.python_extractor.extract_strings(
                force_text(src_data), translatable_file.path[2:]
            )

        # Template file
        return extract_transifex_template_strings(
            src_data, translatable_file.path[2:], encoding,
        )

    def _read_file(self, path, encoding):
        with io.open(path, 'r', encoding=encoding) as fp:
            return fp.read()

    def _find_files(self, root):
        """Get all files in the given root.

        :param basestring root: the root path to search in
        :return: a list of TranslatableFile objects
        :rtype: list
        """
        # TODO: See if we can remove functionality about locale dir and simplify

        def is_ignored(path, ignore_patterns):
            """Check if the given path should be ignored or not."""
            filename = os.path.basename(path)

            def ignore(pattern):
                return fnmatch.fnmatchcase(filename, pattern) or fnmatch.fnmatchcase(
                    path, pattern)

            return any(ignore(pattern) for pattern in ignore_patterns)

        ignore_patterns = [os.path.normcase(p) for p in self.ignore_patterns]
        dir_suffixes = {'%s*' % path_sep for path_sep in {'/', os.sep}}
        norm_patterns = []
        for p in ignore_patterns:
            for dir_suffix in dir_suffixes:
                if p.endswith(dir_suffix):
                    norm_patterns.append(p[:-len(dir_suffix)])
                    break
            else:
                norm_patterns.append(p)

        all_files = []
        ignored_roots = []
        if self.settings_available:
            ignored_roots = [os.path.normpath(p) for p in
                             (settings.MEDIA_ROOT, settings.STATIC_ROOT) if p]

        for dirpath, dirnames, filenames in os.walk(root, topdown=True,
                                                    followlinks=self.symlinks):
            for dirname in dirnames[:]:
                if (is_ignored(os.path.normpath(os.path.join(dirpath, dirname)),
                               norm_patterns) or
                        os.path.join(os.path.abspath(dirpath),
                                     dirname) in ignored_roots):
                    dirnames.remove(dirname)
                    self.verbose('Ignoring directory %s' % dirname)
                elif dirname == 'locale':
                    dirnames.remove(dirname)
                    self.locale_paths.insert(0, os.path.join(os.path.abspath(dirpath),
                                                             dirname))
            for filename in filenames:
                file_path = os.path.normpath(os.path.join(dirpath, filename))
                file_ext = os.path.splitext(filename)[1]
                if file_ext not in self.extensions or is_ignored(file_path,
                                                                 self.ignore_patterns):
                    self.verbose('Ignoring file %s in %s' %
                                 (filename, dirpath))
                else:
                    locale_dir = None
                    for path in self.locale_paths:
                        if os.path.abspath(dirpath).startswith(os.path.dirname(path)):
                            locale_dir = path
                            break
                    if not locale_dir:
                        locale_dir = self.default_locale_path
                    if not locale_dir:
                        locale_dir = NO_LOCALE_DIR
                    all_files.append(TranslatableFile(
                        dirpath, filename, locale_dir))

        return sorted(all_files)

    def _show_collect_results(self):
        """Display results of collecting source strings from files."""
        total_strings = sum([x[1] for x in self.stats['strings']])
        self.output(
            'Processed {} file(s) and found {} translatable strings '
            'in {} of them'.format(
                self.stats['processed_files'], total_strings, len(self.stats)
            )
        )

    def _show_push_results(self, status_code, response_content):
        """Display results of pushing the source strings to CDS.

        :param int status_code: the HTTP status code
        :param dict response_content: the content of the response
        """
        try:
            if 200 <= status_code < 300:
                # {"created":0,"updated":5,"skipped":1,"deleted":0,"failed":0,"errors":[]}
                created = response_content.get('created')
                updated = response_content.get('updated')
                skipped = response_content.get('skipped')
                deleted = response_content.get('deleted')
                failed = response_content.get('failed')
                errors = response_content.get('errors', [])
                self.output(
                    'Successfully pushed strings to Transifex.\n'
                    'Status: {code}\n'
                    'Created strings: {created}\n'
                    'Updated strings: {updated}\n'
                    'Skipped strings: {skipped}\n'
                    'Deleted strings: {deleted}\n'
                    'Failed strings: {failed}\n'
                    'Errors: {errors}\n'.format(
                        code=status_code,
                        created=created,
                        updated=updated,
                        skipped=skipped,
                        deleted=deleted,
                        failed=failed,
                        errors='\n'.join(errors)
                    )
                )
            else:
                message = response_content.get('message')
                details = response_content.get('details')
                self.output(
                    'Could not push strings to Transifex.\n'
                    'Status: {code}\n'
                    'Message: {message}\n'
                    'Details: {details}\n'.format(
                        code=status_code,
                        message=message,
                        details=json.dumps(details, indent=4),
                    )
                )
        except:
            self.output(
                '(Error while printing formatted report, '
                'falling back to raw format)\n'
                'Status: {code}\n'
                'Content: {content}'.format(
                    code=status_code,
                    content=response_content,
                )
            )

    @cached_property
    def settings_available(self):
        try:
            settings.LOCALE_PATHS
        except ImproperlyConfigured:
            if self.verbosity > 1:
                self.stderr.verbose("Running without configured settings.")
            return False
        return True

    def verbose(self, msg):
        if self.verbosity > 1:
            print(msg)

    def output(self, msg):
        print(msg)



import importlib
from datetime import datetime
from hashlib import md5

import pytz
import six

if six.PY3:
    unicode_compat = str
else:
    unicode_compat = unicode


def generate_key(source_string, context=None):
    """Return a unique key based on the given source string and context.

    :param unicode source_string: the source string
    :param Union[unicode, list] context: an optional context that accompanies the string
    :return: a unique key
    :rtype: str
    """
    if context:  # pragma: no cover
        if isinstance(context, list):
            context = u':'.join(context)
        else:
            context = u':'.join(context.split(','))
    if not context:  # pragma: no cover
        context = ''
    return md5(
        (':'.join([source_string, context])).encode('utf-8')
    ).hexdigest()


def now():
    """Return the current datetime, as a UTC-aware object.

    :rtype: datetime
    """
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def import_to_python(import_str):
    """Given a string 'a.b.c' return object c from a.b module.

    :param str import_str: a path like a.b.c.
    """
    mod_name, obj_name = import_str.rsplit('.', 1)
    obj = getattr(importlib.import_module(mod_name), obj_name)
    return obj


def make_hashable(data):
    """Make the given object hashable.

    It makes it ready to use in a `hash()` call, making sure that
    it's always the same for lists and dictionaries if they have the same items.

    :param object data: the object to hash
    :return: a hashable object
    :rtype: object
    """
    if isinstance(data, (list, tuple)):
        return tuple((make_hashable(item) for item in data))
    elif isinstance(data, dict):
        return tuple(
            (key, make_hashable(value))
            for key, value in sorted(data.items())
        )
    else:
        return data



import logging
import sys

import requests
from transifex.native.consts import (KEY_CHARACTER_LIMIT,
                                     KEY_DEVELOPER_COMMENT, KEY_TAGS)

TRANSIFEX_CDS_HOST = 'https://cds.svc.transifex.net'

TRANSIFEX_CDS_URLS = {
    'FETCH_AVAILABLE_LANGUAGES': '/languages',
    'FETCH_TRANSLATIONS_FOR_LANGUAGE': '/content/{language_code}',
    'PUSH_SOURCE_STRINGS': '/content/'
}

logger = logging.getLogger('transifex.native.cds')
logger.addHandler(logging.StreamHandler(sys.stderr))


# A mapping of meta keys
# (interface_key: cds_key)
# Only contains meta keys that are different between the two
MAPPING = {
    KEY_DEVELOPER_COMMENT: 'developer_comment',
    KEY_CHARACTER_LIMIT: 'character_limit',
    KEY_TAGS: 'tags',
}


class CDSHandler(object):
    """Handles communication with the Content Delivery Service."""

    def __init__(self, configured_languages, token, secret=None, host=TRANSIFEX_CDS_HOST):
        """Constructor.

        :param list configured_languages: a list of language codes for the
            configured languages in the application
        :param str token: the API token to use for connecting to the CDS
        :param str host: the host of the Content Delivery Service
        """
        self.configured_language_codes = configured_languages
        self.token = token
        self.secret = secret
        self.host = host or TRANSIFEX_CDS_HOST

    def fetch_languages(self):
        """Fetch the languages defined in the CDS for the specific project.

        :return: a list of language information
        :rtype: dict
        """

        cds_url = TRANSIFEX_CDS_URLS['FETCH_AVAILABLE_LANGUAGES']
        languages = []

        try:
            response = requests.get(
                self.host + cds_url,
                headers={"Authorization": "Bearer {token}".format(
                    token=self.token)}
            )

            if not response.ok:
                logger.error(
                    'Error retrieving languages from CDS: `{}`'.format(
                        response.reason
                    )
                )
                response.raise_for_status()

            json_content = response.json()
            languages = json_content['data']

        except (KeyError, ValueError):
            # Compatibility with python2.7 where `JSONDecodeError` doesn't exist
            logger.error(
                'Error retrieving languages from CDS: Malformed response')
        except requests.ConnectionError:
            logger.error(
                'Error retrieving languages from CDS: ConnectionError')
        except Exception as e:
            logger.error(
                'Error retrieving languages from CDS: UnknownError (`{}`)'.format(
                    str(e)
                )
            )

        return languages

    def fetch_translations(self, language_code=None):
        """Fetch all translations for the given organization/project/(resource)
        associated with the current token.

        :return: a dictionary with all corresponding translations
        :rtype: dict
        """

        cds_url = TRANSIFEX_CDS_URLS['FETCH_TRANSLATIONS_FOR_LANGUAGE']

        translations = {}

        try:

            if not language_code:
                languages = [lang['code'] for lang in self.fetch_languages()]
            else:
                languages = [language_code]

            for language_code in set(languages) & \
                    set(self.configured_language_codes):
                response = requests.get(
                    self.host + cds_url.format(language_code=language_code),
                    headers={"Authorization": "Bearer {token}".format(
                        token=self.token)}
                )

                if not response.ok:
                    logger.error(
                        'Error retrieving translations from CDS: `{}`'.format(
                            response.reason
                        )
                    )
                    response.raise_for_status()

                json_content = response.json()
                translations[language_code] = json_content['data']

        except (KeyError, ValueError):
            # Compatibility with python2.7 where `JSONDecodeError` doesn't exist
            logger.error(
                'Error retrieving translations from CDS: Malformed response')
        except requests.ConnectionError:
            logger.error(
                'Error retrieving translations from CDS: ConnectionError')
        except Exception as e:
            logger.error(
                'Error retrieving translations from CDS: UnknownError (`{}`)'.format(
                    str(e)
                )
            )

        return translations

    def push_source_strings(self, strings, purge=False):
        """Push source strings to CDS.

        :param list(SourceString) strings: a list of `SourceString` objects holding
            source strings
        :param bool purge: True deletes destination source content not included in
                           pushed content.
                           False appends the pushed content to destination source
                           content.
        :return: the HTTP response object
        :rtype: requests.Response
        """
        if not self.secret:
            raise Exception(
                'You need to use `TRANSIFEX_SECRET` when pushing source content'
            )

        cds_url = TRANSIFEX_CDS_URLS['PUSH_SOURCE_STRINGS']

        data = {k: v for k, v in [self._serialize(item) for item in strings]}
        try:
            response = requests.post(
                self.host + cds_url,
                headers={"Authorization": "Bearer {token}:{secret}".format(
                    token=self.token, secret=self.secret)
                },
                json={
                    'data': data,
                    'meta': {'purge': purge},
                }
            )
            response.raise_for_status()

        except requests.ConnectionError:
            logger.error(
                'Error pushing source strings to CDS: ConnectionError')
        except Exception as e:
            logger.error(
                'Error pushing source strings to CDS: UnknownError (`{}`)'.format(
                    str(e)
                )
            )

        return response

    def _serialize(self, source_string):
        """Serialize the given source string to a format suitable for the CDS.

        :param transifex.native.parsing.SourceString source_string: the object
            to serialize
        :return: a tuple that contains ths string key and its data,
            as (key, data)
        :rtype: tuple
        """
        data = {
            'string': source_string.string,
            'meta': {
                MAPPING.get(k, k): v
                for k, v in source_string.meta.items()
            },
        }
        if source_string.context:
            data['meta']['context'] = source_string.context

        return source_string.key, data



from __future__ import unicode_literals

import ast
import json
import re

from six import string_types
from transifex.common.utils import generate_key, make_hashable
from transifex.native import consts
from transifex.native.consts import KEY_CONTEXT

# PEP 263 magic comment for source code encodings
# e.g. "# -*- coding: <encoding name> -*-"

ENCODING_PATTERN = re.compile('#.*coding[:=]\s*utf-?8', re.IGNORECASE)


# A list of the translate functions that TxNative supports,
# in the syntax of module.deeper_module..function
DEFAULT_MODULES = [
    'transifex.native.translate',
]


class SourceString(object):
    """A data object that contains information about a source string."""

    def __init__(self, string, _context=None, **meta):
        """Constructor.

        :param unicode string: the source string
        :param unicode _context: an optional context that accompanies
            the source string
        """
        self.key = generate_key(string, context=_context)
        self.string = string
        self.context = (
            [x.strip() for x in _context.split(',')] if _context
            else None
        )  # type: list
        self.meta = self._transform_meta(meta)

    @property
    def developer_comment(self):
        """An optional developer comment for the string.

        :rtype: unicode
        """
        return self.meta.get(consts.KEY_DEVELOPER_COMMENT)

    @property
    def character_limit(self):
        """An optional character limit for the string or None if not defined.

        :rtype: int
        """
        return self.meta.get(consts.KEY_CHARACTER_LIMIT)

    @property
    def tags(self):
        """A list of tags.

        :rtype: list
        """
        return self.meta.get(consts.KEY_TAGS, [])

    def _transform_meta(self, meta):
        """Transform values in meta object, whenever applicable.

        :param dict meta:
        :return: the same dictionary, with some values potentially altered
        :rtype: dict
        """
        tags = meta.get(consts.KEY_TAGS)
        if tags and isinstance(tags, string_types):
            meta[consts.KEY_TAGS] = [x.strip() for x in tags.split(',')]

        return {
            k: v for k, v in meta.items()
            if k in consts.ALL_KEYS
        }

    def __repr__(self):
        return '<{}: {}>'.format(
            self.__class__.__name__,
            ' '.join((self.context or []) + [self.string]),
        )

    def __eq__(self, other):
        """Object equality.

        :param SourceString other: the instance to compare to
        :return: True if the instances are equal, False otherwise
        :rtype: bool
        """
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((self.key, make_hashable(self.meta)))


class Extractor(object):
    """Extracts translatable source strings from Python files.

    It also keeps stats about any errors that have occurred.

    It allows clients to register custom module/function paths in order
    to support translate modules and functions that exist in framework
    implementations that use native, such as a Django or a Flask SDK.
    """

    def __init__(self):
        self.errors = []
        self._functions = []
        for path in DEFAULT_MODULES:
            self.register_functions(path)

    def register_functions(self, *func_paths):
        """Register a custom function to be detected during extraction.

        Each arg in `func_paths` must be a string, representing the full path
            of the function, like 'module.deeper_module.func_name'
        """
        for func_path in func_paths:
            nodes = func_path.split('.')
            if len(nodes) < 2:
                raise ValueError(
                    'Function path must contain at least a module and a function, '
                    'e.g. "my_module.translate"'
                )
            self._functions.append(
                ('.'.join(nodes[:-1]), nodes[-1])
            )

    def extract_strings(self, src, origin=None):
        """Parse the given Python file string and extract translatable content.

        :param unicode src: a chunk of Python code
        :param str origin: the filename of the code, i.e. the filename it came from
        :return: a list of SourceString objects
        :rtype: list
        """
        # Replace utf-8 magic comment, to avoid getting a
        # "SyntaxError: encoding declaration in Unicode string"
        src = ENCODING_PATTERN.sub('# ', src)
        try:
            tree = ast.parse(src)
            visitor = TransifexVisitor(self._functions)
            visitor.visit(tree)
            return visitor.source_strings
        except Exception as e:
            # Store an exception for this particular file
            self.errors.append((origin, e))
            return []


class TransifexVisitor(ast.NodeVisitor):
    """A visitor subclass that detects calls to TxNative translate methods
    and creates SourceString objects from them.

    Detects the source string, as well as the context and any key/value
    parameters, if they exist.

    For each Python file, a separate instance of TransifexVisitor is created.

    NOTE: in order for a function call to be detected, the corresponding
    `import` statement must appear before in the syntax tree. If we want
    to support calls that appear before imports, we could visit the tree
    twice, once to detect the imports and once to detect the calls
    and create the strings.
    """

    def __init__(self, registered_calls):
        """Constructor.

        Each item in `registered_calls` should be a tuple like:
          ('a.b.c', 'translate_func')
        which translates to calls like:
          from a.b.c import translate_func as tr; tr(...)
          from a.b import c; c.translate_func(...)
        and so on

        :param list registered_calls: a list of 2-tuples each with information
            on how to detect the imports and the calls
        """
        super(TransifexVisitor, self).__init__()
        self.source_strings = []

        # Contains a list of module/function pairs in dot notation
        # that have been registered externally.
        # Example: [('a.b', 'translate'), ('foo', 'trans')]
        self._registered_calls = registered_calls

        # Dynamically populated when parsing import statements,
        # this list will contain the module/function pars that are
        # actually found in the current syntax tree (e.g. in a specific Python file).
        # This way, only calls that are actually made to TxNative-related functions
        # will be matched for each file, instead of matching any function that
        # has a name identical to what TxNative provides (e.g. a `translate()` method
        # of another module).
        # Support 'as' syntax, e.g. import translate as _trans
        # Example: [('a.b', 'translate_string'), ('foo', '_trans')]
        self._supported_calls = []

    def visit_Import(self, node):
        """Support the 'import native as _native' type of imports.

        Given a supported function path that looks like this:
            >>> 'a.b.c.d.translate'
        takes an import statement that looks like this:
            >>> import a.b as _b
        and comes up with a tuple like this:
            >>> ('_b.c.d', 'translate')
        """
        self.generic_visit(node)

        for module_path, func_name in self._registered_calls:
            name = node.names[0].name
            as_name = node.names[0].asname
            # e.g. module_path='a.b.c', func_name='translate',
            #      name='a.b', asname='_c'
            if not module_path.startswith(name):
                continue

            remaining_module_path = module_path.replace(name, '')
            remaining_module_path = remaining_module_path.lstrip('.')

            if not as_name:
                as_name = name.split('.')[-1]

            remaining_module_path = (
                '{}.{}'.format(as_name, remaining_module_path).rstrip('.')
            )

            self._supported_calls.append(
                (remaining_module_path, func_name)
            )

    def visit_ImportFrom(self, node):
        """Support the 'from native import native as _native' type of imports.

        Given a supported function path that looks like this:
            >>> 'a.b.c.d.translate'
        takes an import statement that looks like this:
            >>> from a.b import c as _c
        and comes up with a tuple like this:
            >>> ('_c.d', 'translate')
        """
        self.generic_visit(node)

        # Loop through all registered module/function calls,
        # e.g. transifex.native.django.t
        # and see if the current import matches any of them
        for registered_module_path, registered_func_name in self._registered_calls:
            # e.g. registered_module_path='transifex.native.django',
            #      registered_func_name='t'

            module = node.module
            if not registered_module_path.startswith(module):
                continue

            # Loop through all 'import' statements, e.g.
            # from m import a, b, c -> loop through [a, b, c] objects
            for name_obj in node.names:
                name = name_obj.name
                as_name = name_obj.asname

                try:
                    # The full function call in the code is identical to the
                    # registered function name, e.g. it's `ut('...')`
                    if name == registered_func_name:
                        registered_func_name = as_name or name
                        remaining_module_path = ''
                    else:
                        modules = registered_module_path.split('.')
                        if name in modules:
                            modules = modules[modules.index(name):]
                            if as_name and modules[0] == name:
                                modules[0] = as_name
                            remaining_module_path = '.'.join(modules)
                        else:
                            continue
                except Exception as e:
                    print(
                        'Error while visiting node: {}.{}{}: {}'.format(
                            module, name, (' as ' +
                                           as_name if as_name else ''), e
                        )
                    )
                    continue

                self._supported_calls.append(
                    (remaining_module_path, registered_func_name)
                )

    def visit_Call(self, node):
        """Extract a source string from a "translate" function call.

        Supports calls like:
          >>> translate('...')
          >>> a.b.c.translate('...')
        based on the supported calls that have been detected
        for the current syntax tree.
        """
        self.generic_visit(node)

        # Find the full module/function path of the current calling node,
        # e.g. 'a.b.c.translate'.
        # The way to retrieve the module name and function name
        # differs a lot depending on the level of nesting, so try...catch blocks
        # are used in order to cover all cases
        modules = []
        try:
            current_node = node.func.value
            current_func_name = node.func.attr
            # Module nesting can be indefinite, so we need to follow it
            # to the end
            while True:
                try:
                    modules.insert(0, current_node.attr)
                    current_node = current_node.value
                except AttributeError:
                    modules.insert(0, current_node.id)
                    break
        except AttributeError:
            try:
                current_func_name = node.func.id
            except AttributeError:
                try:
                    current_func_name = node.func.attr
                except AttributeError as e:
                    raise AttributeError(
                        'Invalid module/function format on line {} col {}: {}'.format(
                            node.lineno, node.col_offset, e
                        )
                    )

        current_module_path = '.'.join(modules)

        # Check against all supported function calls and if there is a match
        # create a SourceString with the parsed information
        for module_path, func_name in self._supported_calls:
            if module_path == current_module_path and func_name == current_func_name:
                try:
                    string = node.args[0].s
                    # Context could be passed as an argument, e.g. t('str', 'context')
                    context = node.args[1].s if len(node.args) > 1 else None

                    # Find all custom parameters, e.g. developer comments etc
                    params = {}
                    for keyword in node.keywords:
                        name, value = self._render_keyword(keyword)
                        if value is not None:
                            params[name] = value

                    # If no context was found before, maybe it was passed as a kwarg
                    if context is None:
                        context = params.pop(KEY_CONTEXT, None)

                    source_string = SourceString(string, context, **params)
                    self.source_strings.append(source_string)
                except Exception as e:
                    raise AttributeError(
                        'Invalid module/function format on line {} col {}: {}'.format(
                            node.lineno, node.col_offset, e
                        )
                    )

    def _render_keyword(self, keyword):
        """Render the given keyword to a proper value.

        Processes Keyword objects and returns the keyword name along with
        a value of their corresponding type, i.e. a string, a number or a boolean.

        :param Keyword keyword: the keyword object
        :return: a tuple like (<key>, <value>)
        :rtype: tuple
        """
        if isinstance(keyword.value, ast.Str):
            val = keyword.value.s
        elif isinstance(keyword.value, ast.Num):
            val = keyword.value.n
        elif (isinstance(keyword.value, ast.Name)
              and keyword.value.id in ('True', 'False')):
            val = keyword.value.id == 'True'
        else:
            val = None

        return keyword.arg, val



from __future__ import unicode_literals

from django.template.base import (TOKEN_BLOCK, TOKEN_COMMENT, TOKEN_TEXT,
                                  TOKEN_VAR, TRANSLATOR_COMMENT_MARK,
                                  DebugLexer, Lexer, Parser)
from django.utils.encoding import force_text
from transifex.native.django.templatetags.transifex import \
    parse_translatable_tag

# Django template consts
LOAD_TAG = 'load'
WITH_TAG = 'with'
ENDWITH_TAG = 'endwith'
COMMENT_TAG = 'comment'
ENDCOMMENT_TAG = 'endcomment'
TRANSLATE_TAGS = ['trans', 'translate']
BLOCK_TRANSLATE_TAGS = ['blocktrans', 'blocktranslate']
ENDBLOCK_TRANSLATE_TAGS = ['endblocktrans', 'endblocktranslate']
DJANGO_i18n_TAG_NAME = 'i18n'
TRANSIFEX_TAG_NAME = 'transifex'
COMMENT_FOUND = object()
COPY_AS_IS = object()


def find_filter_identity(filter_name):
    """Return a filter that does no filtering, i.e always returns all items.

    This function supposedly returns a filter object, which in this case
    does nothing other than returning the variable as is.

    :param str filter_name:
    :return: a callable that represents a filter that does not filtering
    :rtype: callable
    """
    return lambda obj: obj


def extract_transifex_template_strings(src, origin=None, charset='utf-8'):
    """Parse the given template and extract translatable content
    based on the syntax supported by Transifex Native.

    Supports the {% t %} template tags.

    :param unicode src: the whole Django template
    :param str origin: an optional context for the filename of the source,
        e.g. the file name
    :param str charset: the character set to use
    :return: a list of SourceString objects
    :rtype: list
    """
    src = force_text(src, charset)
    tokens = Lexer(src).tokenize()
    parser = Parser(tokens, {}, [], origin)
    # Since no template libraries are loaded when this code is running,
    # we need to override the find function in order to use the functionality
    # of the Parser class. The overriden function returns the object as given.
    # Without the override, a KeyError is raised inside the parser.
    parser.find_filter = find_filter_identity

    strings = []
    while parser.tokens:
        token = parser.next_token()
        if token.token_type == TOKEN_BLOCK:
            source_string, _ = parse_translatable_tag(parser, token)
            if source_string is None:
                continue

            strings.append(source_string)

    return strings



# -*- coding: utf-8 -*-
import pytest
from django.template import TemplateSyntaxError
from transifex.native.django.utils.templates import \
    extract_transifex_template_strings
from transifex.native.parsing import SourceString

# A template to use for tests
# The `anyparenttag` tag is added so that we make sure the t/ut tags
# work properly when nested
TEMPLATE = u"""
{% load transifex %}

{% anyparenttag content_main %}
    <p>{{string1}}</p>
    <p>{{string2}}</p>
{% anyparenttag %}
"""


class TestTemplates(object):
    """Test the utils.extract_transifex_template_strings() method."""

    def test_basic(self):
        src = TEMPLATE.replace(
            u'{{string1}}',
            u'{% t "A very important sentence" %}'
        ).replace(
            u'{{string2}}',
            u'{% t "I will look for {total} items" total=bucket.items|length %}'
        )
        strings = extract_transifex_template_strings(src)
        assert strings == [
            SourceString(u"A very important sentence"),
            SourceString(u"I will look for {total} items"),
        ]

    def test_all_string_params(self):
        src = TEMPLATE.replace(
            u'{{string1}}',
            u'{%t "Le canapé" '
            u'_context="furniture" _comment="_comment" _charlimit=10 _tags="t1,t2" %}'
        )
        strings = extract_transifex_template_strings(src)
        assert strings == [
            SourceString(
                u"Le canapé", _context=u'furniture', _comment=u'_comment',
                _charlimit=10, _tags=['t1', 't2']
            ),
        ]

    def test_multiline(self):
        src = TEMPLATE.replace(
            u'{{string1}}',
            u"""
{% t visit_type='first' username=user.name _context="stuff" _charlimit=10 %}
{visit_type, select,
    first {Welcome, {username}}
    returning {Welcome back, {username}}
}
{% endt %}"""
        )
        strings = extract_transifex_template_strings(src)
        assert strings[0] == SourceString(
            u"""{visit_type, select,
    first {Welcome, {username}}
    returning {Welcome back, {username}}
}""",
            _context='stuff',
            _charlimit=10,
        )

    def test_raises_exception_if_matching_end_tag_not_found(self):
        src = TEMPLATE.replace(
            u'{{string1}}',
            u"""
            {% t %}
            Test
            {% endut %}"""  # `endt` is expected instead
        )
        with pytest.raises(TemplateSyntaxError):
            extract_transifex_template_strings(src)

        src = TEMPLATE.replace(
            u'{{string1}}',
            u"""
            {% ut %}
            Test
            {% endt %}"""  # `endut` is expected instead
        )
        with pytest.raises(TemplateSyntaxError):
            extract_transifex_template_strings(src)
